(defpackage :6502 (:use :common-lisp))
(in-package :6502)

;;;; ---------------------------------------------------------------------
;;;; 
;;;;  Hef's 6502 Assembler
;;;; 
;;;;  Note: The AND and BIT instructions are renamed to ANDA and BITA,
;;;;  respectively, so that they do not clash with CL symbols.
;;;;
;;;;  TODO: Rename assemble-parameter, improve context available for error
;;;;        reports.
;;;; ---------------------------------------------------------------------

(defconstant +6502-nmi-vector+   #xFFFA)
(defconstant +6502-reset-vector+ #xFFFC)
(defconstant +6502-irq-vector+   #xFFFE)

;;; Assembly Utilities

(defun lsb-of (x) (ldb (byte 8 0) x))
(defun msb-of (x) (ldb (byte 8 8) x))

(defun encode-8-bit (x)
  (unless (and (integerp x)
               (>= x -128)
               (< x 256))
    (error "Operand ~A cannot be encoded as an 8-bit operand." x))
  (vector (lsb-of x)))

(defun encode-16-bit (x)
  (unless (and (integerp x)
               (>= x 0)
               (< x 65536))
    (error "Operand ~A cannot be encoded as a 16-bit operand." x))
  (vector (lsb-of x) (msb-of x)))

(defun orf (x y)
  (unless (zerop (logand x y))
    (error "Bitmasks ~A and ~A overlap!" x y))
  (logior x y))

(defun dumpbin (filename vector)
  (with-open-file (out filename :if-exists :supersede
                                :direction :output
                                :element-type '(unsigned-byte 8))
    (write-sequence vector out)))

(defun invalid-operand-error (instr-description operand)
  (error "Invalid operand or addressing mode for ~A: ~A"
         (or instr-description "this instruction")
         operand))

;;;; Addressing Mode Definitions

(defclass 6502-addressing-mode () ())
(defclass 6502-amode-with-param (6502-addressing-mode)
  ((parameter :reader parameter :initarg :parameter)))
(defclass 6502-amode-param-8  (6502-amode-with-param) ())
(defclass 6502-amode-param-16 (6502-amode-with-param) ())

;; def-addr-mode: Macro to automate generation of addressing mode classes and
;; constructor functions.
(defmacro def-addr-mode (name superclass-list)
  `(progn
    (defclass ,name ,superclass-list ())
    (defgeneric ,name (param))
    (defmethod ,name (param) (make-instance ',name :parameter param))))

;; Implicit address modes are currently specified by passing nil to #'assemble
;; We consider the accumulator address mode a special case of implicit addressing.

(defclass zero-page-mode (6502-amode-param-8) ()) ;; Expressions containing an address in the zero page
(defclass absolute-mode (6502-amode-param-16) ()) ;; Expressions containing a 16-bit literal address

(def-addr-mode imm  (6502-amode-param-8))  ;; Immediate
(def-addr-mode zp   (zero-page-mode))      ;; Zero Page
(def-addr-mode zpx  (zero-page-mode))      ;; Zero Page, X
(def-addr-mode zpy  (zero-page-mode))      ;; Zero Page, Y {for STX/LDX instructions}
(def-addr-mode idxi (zero-page-mode))      ;; Indexed Indirect ($aa,X)
(def-addr-mode indi (zero-page-mode))      ;; Indirect Indexed ($aa),Y
(def-addr-mode mem  (absolute-mode))       ;; Absolute Address
(def-addr-mode abx  (absolute-mode))       ;; Absolute, X
(def-addr-mode aby  (absolute-mode))       ;; Absolute, Y {for LDX instruction}
(def-addr-mode indirect (absolute-mode))   ;; Indirect
(def-addr-mode rel  (6502-amode-param-8))  ;; PC-Relative offset (for branch instructions)

;; Encoding of parameters, according to instruction addressing mode
(defgeneric assemble-parameter (x))
(defgeneric assemble-operand (expr operand))
(defmethod assemble-parameter ((x 6502-amode-with-param)) (assemble-operand x (parameter x)))
(defmethod assemble-parameter ((x null)) #())  ; Implicit/accumulator operand
(defmethod assemble-operand ((expr 6502-amode-param-8)  (x integer)) (encode-8-bit x))
(defmethod assemble-operand ((expr 6502-amode-param-16) (x integer)) (encode-16-bit x))
(defmethod assemble-operand ((expr indirect) (x integer))
  (if (= #xFF (logand x #xFF))
      (error "Indirect jump through ~A tickles 6502 page wraparound bug, please twiddle your code." x)
      (encode-16-bit x)))

;;;; The 6502 Instruction Set (see http://axis.llx.com/~nparker/a2/opcodes.html)

(defclass 6502-instruction () ())

(defgeneric assemble (instruction parameter))        ; => byte vector
(defgeneric assemble-opcode (instruction parameter)) ; => opcode byte

(defmethod assemble (instruction parameter)
  (error "Don't know how to assemble instruction ~A ~A" instruction parameter))
(defmethod assemble :around (instruction (parameter integer)) ; Syntactic sugar
  (assemble instruction (imm parameter)))
(defmethod assemble ((instruction symbol) parameter)
  (concatenate 'vector (vector (assemble-opcode instruction parameter)) (assemble-parameter parameter)))
(defmethod assemble-opcode (instruction parameter)
  (error "Invalid addressing mode or instruction (~A,~A)" instruction parameter))

;; Group 1:
;;        ORA     AND     EOR     ADC     STA     LDA     CMP     SBC
;; (zp,X)  01      21      41      61      81      A1      C1      E1
;; zp      05      25      45      65      85      A5      C5      E5
;; #       09      29      49      69              A9      C9      E9
;; abs     0D      2D      4D      6D      8D      AD      CD      ED
;; (zp),Y  11      31      51      71      91      B1      D1      F1
;; zp,X    15      35      55      75      95      B5      D5      F5
;; abs,Y   19      39      59      79      99      B9      D9      F9
;; abs,X   1D      3D      5D      7D      9D      BD      DD      FD

(defmethod assemble-opcode ((instruction (eql 'sta)) (operand imm))
  (invalid-operand-error instruction operand))

(defun group-1-addr-code (x)
  (typecase x
    (idxi #b000)  ;   (zero page,X)
    (zp   #b001)  ;   zero page
    (imm  #b010)  ;   #immediate
    (mem  #b011)  ;   absolute
    (indi #b100)  ;   (zero page),Y
    (zpx  #b101)  ;   zero page,X
    (abx  #b110)  ;   absolute,Y
    (aby  #b111)  ;   absolute,X
    (t (invalid-operand-error nil x))))

(defun group-1-asm (parameter opcode)
  (orf
   (orf (ash opcode 5)
        (ash (group-1-addr-code parameter) 2))
   #b01))

(defmacro def6502 (name encoder &rest args)
  `(progn
    (defmethod assemble-opcode ((instruction (eql ',name)) parameter)
      (funcall #',encoder parameter ,@args))
    (defun ,name (&optional operand) (assemble ',name operand))))

(def6502 ORA  group-1-asm #b000)
(def6502 ANDA group-1-asm #b001)
(def6502 EOR  group-1-asm #b010)
(def6502 ADC  group-1-asm #b011)
(def6502 STA  group-1-asm #b100)
(def6502 LDA  group-1-asm #b101)
(def6502 CMP  group-1-asm #b110)
(def6502 SBC  group-1-asm #b111)

;; Group 2:
;;                 ASL     ROL     LSR     ROR     STX     LDX     DEC     INC
;; #                                                       A2               
;; zp              06      26      46      66      86      A6      C6      E6
;; A               0A      2A      4A      6A                               
;; abs             0E      2E      4E      6E      8E      AE      CE      EE
;; zp,X/zp,Y       16      36      56      76      96      B6      D6      F6
;; abs,X/abs,Y     1E      3E      5E      7E              BE      DE      FE

(defun group-2/3-addr-code (x types)
  (unless (typep x types) (invalid-operand-error nil x))
  (typecase x
    (imm  #b000)   ; #immediate
    (zp   #b001)   ; zero page
    (null #b010)   ; accumulator
    (mem  #b011)   ; absolute
    (zpx  #b101)   ; zero page,X
    (zpy  #b101)   ; zero page,Y {for STX, LDX}
    (abx  #b111)   ; absolute,X
    (aby  #b111))) ; absolute,Y  {for LDX}

(defun group-2-asm (parameter opcode &optional (types '(or zp null mem zpx abx)))
  (orf
   (orf (ash opcode 5)
        (ash (group-2/3-addr-code parameter types) 2))
   #b10))

;; Default set of address modes is suitable for ASL, ROR, LSR, ROR
(def6502 ASL group-2-asm #b000)
(def6502 ROL group-2-asm #b001)
(def6502 LSR group-2-asm #b010)
(def6502 ROR group-2-asm #b011)
(def6502 STX group-2-asm #b100 '(or zp mem zpy))
(def6502 LDX group-2-asm #b101 '(or imm zp mem zpy aby))
(def6502 DEC group-2-asm #b110 '(or zp mem zpx abx))
(def6502 INC group-2-asm #b111 '(or zp mem zpx abx))

;;; Group 3:
;;         BIT     JMP     JMP()   STY     LDY     CPY     CPX
;; #                                       A0      C0      E0
;; zp      24                      84      A4      C4      E4
;; abs     2C      4C      6C      8C      AC      CC      EC
;; zp,X                            94      B4               
;; abs,X                                   BC

(defun group-3-asm (parameter opcode types)
  (orf
   (orf (ash opcode 5)
        (ash (group-2/3-addr-code parameter types) 2))
   #b00))

(def6502 BITA group-3-asm #b001 '(or zp mem))         
(def6502 STY  group-3-asm #b100 '(or zp mem zpx))         
(def6502 LDY  group-3-asm #b101 '(or imm zp mem zpx abx))         
(def6502 CPY  group-3-asm #b110 '(or imm zp mem))         
(def6502 CPX  group-3-asm #b111 '(or imm zp mem))

;;; Special case JMP, because the high bits are not fixed.

(defun asm-jmp (parameter)
  (typecase parameter
    (mem      #x4C)
    (indirect #x6C)
    (t (invalid-operand-error 'jmp parameter))))

(def6502 JMP asm-jmp)

;;; Conditional Branches:

(defun simple-instruction (operand value &optional (type 'null))
  (unless (typep operand type) (invalid-operand-error "simple instruction" operand))
  value)

;; The conditional branch instructions all have the form xxy10000. The flag
;; indicated by xx is compared with y, and the branch is taken if they are equal.

;; xx	flag
;; 00	negative
;; 01	overflow
;; 10	carry
;; 11	zero

;; This gives the following branches:
(def6502 BPL simple-instruction #x10 '(or imm rel))
(def6502 BMI simple-instruction #x30 '(or imm rel))
(def6502 BVC simple-instruction #x50 '(or imm rel))
(def6502 BVS simple-instruction #x70 '(or imm rel))
(def6502 BCC simple-instruction #x90 '(or imm rel))
(def6502 BCS simple-instruction #xB0 '(or imm rel))
(def6502 BNE simple-instruction #xD0 '(or imm rel))
(def6502 BEQ simple-instruction #xF0 '(or imm rel))

(defvar *branch-instructions* '(BPL BMI BVC BVS BCC BCS BNE BEQ))

;;; Miscellaneous Instructions:

;; The remaining instructions are probably best considered simply by listing
;; them. Here are the interrupt and subroutine instructions:

(def6502 BRK simple-instruction #x00)
(def6502 JSR simple-instruction #x20 'mem)
(def6502 RTI simple-instruction #x40)
(def6502 RTS simple-instruction #x60)

;; (JSR is the only absolute-addressing instruction that doesn't fit the aaabbbcc pattern.)

;; Other single-byte instructions:

;; PHP 	PLP 	PHA 	PLA 	DEY 	TAY 	INY 	INX
;; 08 	28 	48 	68 	88 	A8 	C8 	E8

;; CLC 	SEC 	CLI 	SEI 	TYA 	CLV 	CLD 	SED
;; 18 	38 	58 	78 	98 	B8 	D8 	F8

;; TXA 	TXS 	TAX 	TSX 	DEX 	NOP
;; 8A 	9A 	AA 	BA 	CA 	EA

(def6502 PHP simple-instruction #x08)
(def6502 PLP simple-instruction #x28)
(def6502 PHA simple-instruction #x48)
(def6502 PLA simple-instruction #x68)
(def6502 DEY simple-instruction #x88)
(def6502 TAY simple-instruction #xA8)
(def6502 INY simple-instruction #xC8)
(def6502 INX simple-instruction #xE8)

(def6502 CLC simple-instruction #x18)
(def6502 SEC simple-instruction #x38)
(def6502 CLI simple-instruction #x58)
(def6502 SEI simple-instruction #x78)
(def6502 TYA simple-instruction #x98)
(def6502 CLV simple-instruction #xB8)
(def6502 CLD simple-instruction #xD8)
(def6502 SED simple-instruction #xF8)

(def6502 TXA simple-instruction #x8A)
(def6502 TXS simple-instruction #x9A)
(def6502 TAX simple-instruction #xAA)
(def6502 TSX simple-instruction #xBA)
(def6502 DEX simple-instruction #xCA)
(def6502 NOP simple-instruction #xEA)

;;;; Single-pass Assembler

;;; A protocol for resolving virtual addresses (in the case of bank-switching)
;;; to 16-bit bus addresses. This is not fully baked..

(defgeneric resolve-address (address))
(defgeneric address-incr (address delta))
(defmethod resolve-address ((address integer)) address)
(defmethod address-incr ((address integer) delta) (+ address delta))

;;; Context Protocol
(defvar *context* nil)
(defgeneric origin (context))
(defgeneric context-get-label (context label))
(defgeneric context-set-label (context label address))
(defgeneric context-emit (context vector))

;;; Integration of Assembly Context

(defmethod context-emit ((context null) vector) vector)

(defmethod assemble :around (instruction parameter)
  (let ((output (call-next-method)))
    (emit output)
    output))

;;; Shorthands

(define-symbol-macro *origin* (origin *context*))
(defun emit (vector) (context-emit *context* vector))
(defun get-label (label) (context-get-label *context* label))
(defun set-label (label &optional address)
  (when *context* (context-set-label *context* label (or address *origin*))))
(defun define-label (label &optional (address *origin*))
  (when (get-label label) (error "Cannot redefine label ~A" label))
  (context-set-label *context* label address))

;; The Basic Assembly Context

(defclass basic-assembly-context ()
  ((origin :accessor origin :initarg :origin :initform nil)
   (symbol-table  :accessor symbol-table :initarg :symbol-table :initform nil)
   (output-vector :accessor output-vector :initform (make-array 0 :adjustable t :fill-pointer t))))

(defmethod context-get-label ((context basic-assembly-context) label)
  (cdr (assoc label (symbol-table context))))

(defmethod context-set-label ((context basic-assembly-context) label address)
  (setf (symbol-table context) (acons label address (symbol-table context))))

(defmethod context-emit ((context basic-assembly-context) (vector vector))
  (map nil (lambda (x) (vector-push-extend x (output-vector context))) vector)
  (incf (origin context) (length vector)))

;;; Fixup Generation (this is getting messy..)

(defun make-fixup (fn &optional (transformer 'identity))
  "If (fn) returns non-nil, return (transformer (fn)), else return a lambda
 which repeats this computation."  
  (if (not (functionp fn))
      (funcall transformer fn)
      (let ((tmp (funcall fn)))
        ;(hef:debugf "make-fixup, " tmp)
        (or (and tmp (funcall transformer tmp)) ; NIL is not a valid result, ever. =/
            (lambda ()
              (let ((x (funcall fn)))
                ;(hef:debugf "attemping to resolve, " x)
                (and x (funcall transformer x))))))))

(defmacro fixup ((var expression) &body body)
  `(make-fixup
    (lambda () (let ((tmp ,expression))
                 ;(format *trace-output* "~&  FIXUP ~A=~A => ~A~%" ',var ',expression tmp)
                 (if (functionp tmp) (funcall tmp) tmp)))
    (lambda (,var) ,@body)))

(defun make-fixup-word (fn transformer)
  (vector
   (fixup (x (funcall fn)) (elt (funcall transformer x) 0))
   (fixup (x (funcall fn)) (elt (funcall transformer x) 1))))
;   (make-fixup (lambda () (funcall fn)) (lambda (x) (elt (funcall transformer x) 0)))
;   (make-fixup (lambda () (funcall fn)) (lambda (x) (elt (funcall transformer x) 1)))))

(defun resolve-fixup (fixup)
  (if (functionp fixup)
      (or (funcall fixup) fixup)
      fixup))



;;; The "Incremental Linker"

(defun link (vector) (map 'vector #'resolve-fixup vector))

;;; The Assembler

(defmacro assembling (label-name (&optional ctx-class &rest ctx-args) &body body)
  (let ((ctx-constructor (if ctx-class
                             (cons `',ctx-class ctx-args)
                             '('basic-assembly-context
                               :origin (origin *context*)
                               :symbol-table (symbol-table *context*)))))
    `(progn
      (set-label ',(or label-name (gensym "BLOCK")))
      (let ((output (let ((*context* (make-instance ,@ctx-constructor)))
                     ,@body
                     (link (output-vector *context*)))))
        (when *context*
          (emit output)
          (format *trace-output* "~&Emitting ~A bytes to context ~A~%"
                  (length output) *context*))
        (values output  *context*)))))

;; Assembler Directives

(defun advance-to (offset &optional (fill-byte 0))
  (let ((delta (- offset (resolve-address *origin*))))
    (when (< offset 0) (error "Cannot advance to ~X, it is less than the current assembly address (~X)"
                              offset (resolve-address *origin*)))
    (emit (make-array delta :initial-element fill-byte))))

(defun align (alignment)
  (format *trace-output* "~&Align by ~A from ~X: Advancing to ~X~%" alignment *origin* (* alignment (ceiling (resolve-address *origin*) alignment)))
  (advance-to (* alignment (ceiling (resolve-address *origin*) alignment))))

(defun db (&rest bytes)  
  (emit (map 'vector (lambda (b)
                       (let ((x (etypecase b (character (char-code b)) (integer b))))
                         (assert (and (>= x 0) (< x 256)))))
             bytes)))

(defun dw (&rest words)
  (map nil (lambda (w)
             (etypecase w
               (integer (assert (and (>= w (- #x8000)) (< w #x10000)))
                        (emit (encode-16-bit w)))
               (symbol (emit (make-fixup-word (lambda () (get-label w)) #'encode-16-bit)))))
       words))

;;; Labels

;; Labels in the Zero Page
(defmethod assemble-operand ((expr zero-page-mode) (operand symbol))
  (vector (make-fixup (lambda ()
                        (let ((address (get-label operand)))
                          (when (and (integerp address) (>= address #x100))
                            (error "Address (~X) of symbol ~A is not in the zero-page."
                                   address operand))
                          address)))))

;; 16-bit Absolute Addresses
(defmethod assemble-operand ((expr absolute-mode) (operand symbol))
  (make-fixup-word (lambda () (get-label operand)) (lambda (x) (assemble-operand expr x))))

;; Syntactic Sugar - Convert expressions consisting of a quoted symbol to
;; absolute or relative addresses, according to instruction type.
(defmethod assemble :around ((instruction symbol) (expression symbol))
  (if (null expression)   ; 'nil' is also a symbol, guard against that.
      (call-next-method)
      (if (member instruction *branch-instructions*)          
          (assemble instruction (rel expression))
          (assemble instruction (mem expression)))))

;; The definitions above don't need these, but they're handy for bashing
;; forward references into unusual places.
(defmethod assemble-operand ((expr 6502-amode-param-8) (operand function))
  (vector (resolve-fixup operand)))

(defmethod assemble-operand ((expr 6502-amode-param-16) (operand function))
  (make-fixup-word operand #'encode-16-bit))

(defmethod assemble-operand ((expr 6502-amode-param-16) (operand 6502-amode-param-16))
  (assemble-operand expr (parameter operand)))

;; Branching to Labels

(defun compute-branch-offset (origin address)
  (let ((delta (- address 2 (resolve-address origin)))) ;; FIXME: don't subtract in integer domain..
    (unless (and (>= delta -128) (< delta 127))
      (error "Branch offset out of range"))
    (logand delta #xFF)))

(defmethod assemble-operand ((expr rel) (operand symbol))
  (let ((origin *origin*))
    (vector (make-fixup (lambda ()
                          (let ((address (get-label operand)))
                            (and address
                                 (compute-branch-offset origin address))))))))

;; Programmer Aids

(defun label-lsb (label) (imm (make-fixup (lambda () (get-label label)) #'lsb-of)))
(defun label-msb (label) (imm (make-fixup (lambda () (get-label label)) #'msb-of)))

;;; Bogus Test Code
#+nil
(dumpbin "/tmp/test.bin"
  (assembling nil (basic-assembly-context :origin #x8000)
    (lda 42)
    (align 16)
    (jsr 'loopfoo)
;    (advance-to #x80FF)  ;; Test indirect mode bug checking
    (set-label 'loopfoo)
    (jmp (mem 'loopfoo))
    (jmp 'loopfoo)
    (jmp 'forward-label)
    (bne (rel 'loopfoo))
    (bne (rel 'forward-label))
    (jmp (indirect 'loopfoo))
;    (lda (zpx 'loopfoo))
    (set-label 'forward-label)
    (rti)
    (jmp (mem *origin*))))




